//
// Exemple d'utilisation simple de la bibliothèque MultipleDistpach
// Auteur : Julien Harbulot
//

#include <iostream>

#include "cpp11compatibility.h" //delete if c++1y
#include "multiple_dispatch.h"

using namespace std;

//===========================================================
// Utilisation basique de la bibliothèque MultipleDispatch.
//
//
// Dans cet exemple, nous allons effectuer un double dispatch
//
// On possède une collection d' Animal* (Chat*, Chien*)
// On veut qu'il puissent s'attaquer entre eux en ayant des
//   comportements différents selon le type de l'attaquant
//   et de l'attaqué.
//
// Pour cela, on définit les comportements voulus dans une
//   classe ActionAttaquer.
// Par exemple : ActionAttaquer::operator()(Chat*, Chat*){}
//
// Ensuite on crée une fonction (ou une méthode) qui prend
//  en argument deux Animal* puis qui appelle le dispatcher,
//  et enfin l'action sur les types retrouvés.
//
// Pour pouvoir retrouver les bons types, il faut indiquer
// à la bibliotheque sur quelle familles on veut travailler.
//
// Une famille est constituée :
// - d'une interface de Visiteur (ex : AnimalVisitor)
// - d'une classe Mère (Animal)
// - de types dérivés (Chat, Chien)
//
// Par exemple, pour créer la famille Animaux :
//
// using Animaux
// = dispatch::Family<AnimalVisitor, Animal, Chat, Chien>;



//===========================================================
// Dans cette partie, tout est classique :
//
// On crée une hierarchie classique (avec pattern Visitor)
//    Chat et Chien héritent de Animal
//    Les classes implémentent : void accept(AnimalVisitor&)
//

class Animal;
class Chat;
class Chien;

class AnimalVisitor{
public:
    virtual void visit(Chat* c) = 0;
    virtual void visit(Chien* c) = 0;
};

class Animal{
public:
    virtual void accept(AnimalVisitor* v) = 0;
    virtual ~Animal(){}
};

class Chat : public Animal{
public:
    virtual void accept(AnimalVisitor* v){ v->visit(this); }
};

class Chien : public Animal{
public:
    virtual void accept(AnimalVisitor* v){ v->visit(this); }
};

//===========================================================
// Ici on utilise la bibliothèque :
//
// On crée une action qui nécessite un multiple dispatch
// On l'enregistre auprès d'un multiple dispatcher
// Nombre de lignes pour utiliser la bibliotheque : 3 lignes.

class ActionAttaquer{
public:
    void operator()(Chat* lhs, Chat* rhs) { cout <<  "griffe - griffe" << endl;}
    void operator()(Chat* lhs, Chien* rhs){ cout <<  "griffe - mord  " << endl;}
    void operator()(Chien* lhs, Chat* rhs){ cout <<  "mord   - griffe" << endl;}
    void operator()(Chien* lhs, Chien* rhs){ cout << "mord   - mord  " << endl;}

    void operator()(Animal* a1, Animal* a2){
        // D'abord on précise sur qui l'on veut effectuer le dispatch
        using Animaux = dispatch::Family<AnimalVisitor, Animal, Chat, Chien>;
        
        // Puis on appelle le Dispatcher
        using Dispatcher = dispatch::Dispatcher<ActionAttaquer, Animaux, Animaux>;
        Dispatcher(this, a1, a2)();
    }
};

//===========================================================
// fonction main(), on crée des Animal* et on essaie le
// double dispatch.

int main()
{
    cout << "Debut." << endl;
    
    Chat chat;
    Chien chien;
    
    Animal* a1 = &chat;
    Animal* a2 = &chien;
    
    ActionAttaquer attaquer;
    attaquer(a1, a2);
    attaquer(a2, a2);
    
    cout << "Fin." << endl;
    return 0;
}

