/*
 Présentation de la bibliothèque Multiple Dispatch Wrapper (partie 1)
 
 Dans ce fichier, nous allons présenter le pattern Visitor,
 puis implémenter un système de double dispatch de deux facons différentes.
 La première présente un double dispatch classique utilisant le pattern Visitor deux fois.
 La seconde utilise la bibliothèque pour simplifier le code et n'utiliser qu'une fois le pattern Visitor.
 */

#include <iostream>
#include "cpp11compatibility.h"
#include "multiple_dispatch.h"
using namespace std;

/*
 Section 0 : Pourquoi utiliser le pattern Visitor ?
 -> Les fonctions n'ont pas de comportement polymorphique en leurs arguments.
 */

namespace introduction {
    
    //
    // Une hiérarchie quelconque.
    //
    class Base{
    public:
        virtual ~Base(){}
        /***/
    };
    
    class Derived1 : public Base{
    public:
        /***/
    };
    
    class Derived2 : public Base{
    public:
        /***/
    };
    
    //
    // Une classe qui opère sur cette hiérarchie.
    //
    class NouveauComportement1{
    public:
        void faire_agir(Base* b){
            cout << "Wrong function." << endl;
        }
        void faire_agir(Derived1* d){
            cout << "Derived1 does something new." << endl;
        }
        void faire_agir(Derived2* d){
            cout << "Derived2 does something new." << endl;
        }
    };
    
    void demonstration(){
        cout << "Démonstration du problème : " << endl;
        Base* b1[] = {new Derived1, new Derived2};
        NouveauComportement1 action;
        for(int i = 0; i < 2; ++i){
            action.faire_agir(b1[i]);
        }
        cout << endl;
        
        for(int i = 0; i < 2; ++i){
            delete b1[i];
        }
    }
    
    /*
     Le problème ici est que faire_agir(Base*) est appelée
     au lieu du faire_agir(Derived1*), comme on l'aurait voulu.
     */
}
/*
 Section 1 : Le pattern Visitor, démonstration
 Le pattern Visitor permet d'ajouter de nouveaux comportements à une hiérarchie existante.
 */

namespace patternVisitor {
    
    //
    // Une hiérarchie qui implémente le pattern Visitor.
    //
    class Base;
    class Derived1;
    class Derived2;
    
    class VisitorBase{
    public:
        virtual void visit(Derived1*) = 0;
        virtual void visit(Derived2*) = 0;
    };
    
    class Base{
    public:
        virtual void accept(VisitorBase* v) = 0;
        virtual ~Base(){}
    };
    
    class Derived1 : public Base{
    public:
        virtual void accept(VisitorBase* v){v->visit(this);}
    };
    
    class Derived2 : public Base{
    public:
        virtual void accept(VisitorBase* v){v->visit(this);}
    };
    
    //
    // Exemple d'utilisation
    //
    
    class NouveauComportement1 : public VisitorBase{
    public:
        void faire_agir(Base* b){
            b->accept(this);
        }
        void visit(Derived1* d){
            cout << "Derived1 does something new ! " << endl;
        }
        void visit(Derived2* d){
            cout << "Derived2 does something new ! " << endl;
        }
    };
    
    void demonstration(){
        cout << "Démonstration du pattern Visitor : " << endl;
        Base* b1[] = {new Derived1, new Derived2};
        NouveauComportement1 action;
        for(int i = 0; i < 2; ++i){
            action.faire_agir(b1[i]);
        }
        cout << endl;
        
        for(int i = 0; i < 2; ++i){
            delete b1[i];
        }
    }
    
    /*
     Grâce à ce procédé les fonctions peuvent tirer profit du polymorphisme des méthodes virtuelles.
     */
}
/*
 Méthode 1 :
 Nous allons maintenant voir comment se servir d'un tel procédé pour effectuer un double dispatch.
 La facon classique pour implémenter un double dispatch est d'utiliser deux fois le pattern Visitor :
 Chaque objet est visitable et visiteur à la fois.
 */

namespace methode1{
    
    //
    // Une hiérarchie qui implémente le pattern Visitor deux fois :
    // chaque objet peut visiter les autres.
    //
    class Base;
    class Derived1;
    class Derived2;
    
    class VisitorBase{
    public:
        virtual void visit(Base*) = 0;
        virtual void visit(Derived1*) = 0;
        virtual void visit(Derived2*) = 0;
        virtual ~VisitorBase(){}
    };
    
    class Base : public VisitorBase{
    public:
        virtual void visit(Base* b){ b->accept(this);}
        virtual void accept(VisitorBase* v) = 0;
        virtual ~Base(){}
    };
    
    class Derived1 : public Base{
    public:
        virtual void accept(VisitorBase* v){v->visit(this);}
        void visit(Derived1* d){
            cout << "Derived1 meets another Derived1" << endl;
        }
        void visit(Derived2* d){
            cout << "Derived1 meets a Derived2" << endl;
        }
    };
    
    class Derived2 : public Base{
    public:
        virtual void accept(VisitorBase* v){v->visit(this);}
        void visit(Derived1* d){
            cout << "Derived2 meets a Derived1" << endl;
        }
        void visit(Derived2* d){
            cout << "Derived2 meets another Derived2" << endl;
        }
    };
    
    //
    // Exemple d'utilisation
    //
    
    
    void demonstration(){
        cout << "Démonstration du pattern Visitor double : " << endl;
        Base* b1[] = {new Derived1, new Derived2};
        
        for(int i = 0; i < 2; ++i){
            b1[i]->visit( b1[(i+1)%2] );
        }
        cout << endl;
        
        for(int i = 0; i < 2; ++i){
            delete b1[i];
        }
    }
    
    /*
     L'inconvénient de cette méthode est que les classes sont surchargées
     par du code dédié au pattern Visitor.
    
     Une telle utilisation soulève de plus des questions d'organisation du code.
     En effet, si le comportement lorsque D1 rencontre D2 est le même
     que celui lorsque D2 rencontre D1, faut-il mettre le code dans D1, dans D2 ou le dupliquer ?
    
     De plus, il est fastidieux d'implémenter deux comportements différents.
     
     Enfin, s'il est possible d'implémenter cette solution pour un double dispatch,
     il serait beaucoup trop fastidieux de le faire pour un triple dispatch ou plus !
    */
}

/*
 Méthode 2 :
 Nous allons maintenant voir comment utiliser la bibliothèque Multiple Dispatch Wrapper
 afin de simplifier la mise en oeuvre du multiple dispatch.
 */
namespace library {
    //
    // Création d'une hiérarchie classique qui implémente un pattern Visitor simple.
    //
    
    class Base;
    class Derived1;
    class Derived2;
    
    class VisitorBase{
    public:
        virtual void visit(Derived1*) = 0;
        virtual void visit(Derived2*) = 0;
    };
    
    class Base{
    public:
        virtual void accept(VisitorBase* v) = 0;
        virtual ~Base(){}
    };
    
    class Derived1 : public Base{
    public:
        virtual void accept(VisitorBase* v){v->visit(this);}
    };
    
    class Derived2 : public Base{
    public:
        virtual void accept(VisitorBase* v){v->visit(this);}
    };

    
    //
    // Création d'un nouveau comportement qui nécéssite un double dispatch,
    // et exemple d'utilisation.
    //
    
    class NouveauComportement2{
    public:
        void operator()( Base* b1, Base* b2){ // Trois lignes seulement !
            using Famille = dispatch::Family<VisitorBase, Base, Derived1, Derived2>;
            using Dispatcher = dispatch::Dispatcher<NouveauComportement2, Famille, Famille>;
            Dispatcher(this, b1, b2)();
        }
        
        void operator()( Derived1* lhs, Derived1* rhs){
            cout << "Derived1 meets another Derived1" << endl;
        }
        void operator()( Derived1* lhs, Derived2* rhs){
            cout << "Derived1 meets a Derived2" << endl;
        }
        void operator()( Derived2* lhs, Derived1* rhs){
            cout << "Derived2 meets a Derived1" << endl;
        }
        void operator()( Derived2* lhs, Derived2* rhs){
            cout << "Derived2 meets another Derived2" << endl;
        }
    };
    
    void demonstration(){
        cout << "Démonstration avec la bibliothèque : " << endl;
        Base* b1[] = {new Derived1, new Derived2};
        
        NouveauComportement2 action;
        for(int i = 0; i < 2; ++i){
            action(b1[i] , b1[(i+1)%2]);
        }
        cout << endl;
        
        for(int i = 0; i < 2; ++i){
            delete b1[i];
        }
    }
    
    /*
     Conclusion :
     L'utilisation de la bibliothèque est simple et peu intrusive :
     le code n'est pas disséminé dans les hiérarchies,
     trois lignes seulement sont ajoutées au début de l'appel.
     
     Enfin, elle permet d'utiliser plusieurs comportements et d'étendre
     le dispatch à plus de 2 arguments (2 familles) tout aussi simplement.
     */
}

/* Exécution des démonstations */

int main(){
    introduction::demonstration();
    patternVisitor::demonstration();
    methode1::demonstration();
    library::demonstration();
}